/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.IFahrtRichtung;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.ISobStrasse;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.IStoerfallIndikatoren;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.StoerfallIndikatorVerwaltung;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterListener;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterObserver;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.SobParameterObserver;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau;
import de.bsvrz.ibv.anastau.progstau.IStauverlaufsPrognose;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modul zur Stauobjektbestimmung.<br>
 * Eine Instanz der Klasse ermittelt, verwaltet und prognostiziert die
 * Stauobjekte für ein VerkehrsmodellNetz.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class StauobjektBestimmung implements IStauobjektBestimmung, ISobParameterListener {

	/**
	 * der Logger für Debug-Ausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Standard-Zykluszeit für die Berechnung der Stauobjekte in
	 * Millisekunden.
	 */
	private static final long DEFAULT_ZYKLUS = 60000L;

	/**
	 * die Verwaltung für die Parameter der Stauobjektbestimmung.
	 */
	private ISobParameterObserver parameter;

	/**
	 * die Verwaltung der Stauindikatoren für diese Instanz der
	 * Stauobjektbestimmung.
	 */
	private IStoerfallIndikatoren indikatoren;

	/**
	 * der letzte Ausführungszeitpunkt der Stauobjektbestimmung.
	 */
	private long lastExecutionTime;

	/**
	 * die Zykluszeit für die Ausführung der Stauobjektbestimmung.
	 */
	private long zyklus = StauobjektBestimmung.DEFAULT_ZYKLUS;

	/**
	 * das VerkehrsmodellNetz, für das Stauobjekte ermittelt werden sollen.
	 */
	private VerkehrModellNetz netz;

	/**
	 * das Prognosemodul für die Instanz der Stauobjektbestimmung, d.h. für das
	 * referenzierte VerkehrsmodellNetz.
	 */
	private IStauverlaufsPrognose prognose;

	/**
	 * Konstruktor.
	 */
	public StauobjektBestimmung() {
		berechneLetztenVirtuelleAusfuehrungszeitpunkt();
	}

	/**
	 * ermittelt den letzten Zeitpunkt der Ausführung der Stauobjektbestimmung.
	 */
	private void berechneLetztenVirtuelleAusfuehrungszeitpunkt() {
		final Calendar cal = Calendar.getInstance();
		cal.set(Calendar.HOUR, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		lastExecutionTime = cal.getTimeInMillis();
		final long now = System.currentTimeMillis();
		while (lastExecutionTime < now) {
			lastExecutionTime += zyklus;
		}
		lastExecutionTime -= zyklus;
	}

	/**
	 * führt die Berechnung der Stauobjekte für alle innerhalb des Moduls
	 * verwalteten Straßenobjekte für den gegbenen Zeitpunkt aus.
	 *
	 * @param zeitpunkt
	 *            der Zeitpunkt für den die Berechnung erfolgen soll
	 * @see de.bsvrz.ibv.anastau.bestimmungstau.IStauobjektBestimmung#calculate(long)
	 */
	@Override
	public final void calculate(final long zeitpunkt) {
		StauobjektBestimmung.LOGGER.info("Berechne Stauobjekte");
		indikatoren.sichereSituation();

		// neue gestaute oder nicht mehr gestaute Bereiche auswerten
		for (final ISobStrasse strasse : indikatoren.getStrassen()) {
			StauobjektBestimmung.LOGGER.finest("Berechne Strasse: " + strasse.getStrasse().getName());
			for (final IFahrtRichtung richtung : strasse.getFahrtRichtungen()) {
				richtung.aktualisiereStaus(zeitpunkt);
				richtung.getStaus().sichern(zeitpunkt);
			}
		}
	}

	@Override
	public final void changed(final ISobParameterObserver parameterObserver) {
		StauobjektBestimmung.LOGGER.config(parameterObserver.getParameterString());
		zyklus = parameterObserver.getZyklusZeit() * Constants.MILLIS_PER_SECOND;
		berechneLetztenVirtuelleAusfuehrungszeitpunkt();
	}

	/**
	 * liefert die Indikatorverwaltung des Moduls.
	 *
	 * @return das Modul
	 */
	public final IStoerfallIndikatoren getIndikatoren() {
		return indikatoren;
	}

	/**
	 * liefert die letzte Ausführungszeit.
	 *
	 * @return lastExecutionTime
	 */
	@Override
	public final long getLastExecutionTime() {
		return lastExecutionTime;
	}

	@Override
	public final VerkehrModellNetz getNetz() {
		return netz;
	}

	/**
	 * liefert den nächsten Ausführungszeitpunkt der Stauobjektbestimmung.
	 *
	 * @return nextExecutionTime
	 */
	@Override
	public final long getNextExecutionTime() {
		final long result = lastExecutionTime + zyklus;
		StauobjektBestimmung.LOGGER.finer("Nächster Ausführungszeitpunkt der Stauobjektbestimmung: "
				+ DateFormat.getDateTimeInstance().format(new Date(result)));

		return result;
	}

	@Override
	public final ISobParameterObserver getParameter() {
		return parameter;
	}

	@Override
	public final IStauverlaufsPrognose getPrognoseModul() {
		return prognose;
	}

	/**
	 * liefert die Liste aller Stauobjekte, die innerhalb des Moduls ermittelt
	 * wurden.
	 *
	 * @return die Liste der Stau-Objekte
	 * @see de.bsvrz.ibv.anastau.bestimmungstau.IStauobjektBestimmung#getStauList()
	 */
	@Override
	public final Collection<IStau> getStauList() {
		final Collection<IStau> result = new ArrayList<>();
		for (final ISobStrasse strasse : indikatoren.getStrassen()) {
			for (final IFahrtRichtung richtung : strasse.getFahrtRichtungen()) {
				result.addAll(richtung.getStaus().getStauMenge());
			}
		}
		return result;
	}

	@Override
	public final void init(final ClientDavInterface dav, final VerkehrModellNetz verkehrsNetz,
			final IStauverlaufsPrognose prognoseModul) {
		this.netz = verkehrsNetz;
		this.netz.stausBereinigen(false);
		this.prognose = prognoseModul;
		parameter = new SobParameterObserver(dav, verkehrsNetz);
		parameter.addParameterListener(this);
		indikatoren = new StoerfallIndikatorVerwaltung(parameter, verkehrsNetz);
	}

	/**
	 * setzt die letzte Ausführungszeit der Stauobjektbestimmung.
	 *
	 * @param lastExecutionTime
	 *            Festzulegender lastExecutionTime
	 */
	@Override
	public final void setLastExecutionTime(final long lastExecutionTime) {
		this.lastExecutionTime = Math.min(System.currentTimeMillis(), lastExecutionTime);
		StauobjektBestimmung.LOGGER.finer("Setzte letzte Ausführungszeit: "
				+ DateFormat.getDateTimeInstance().format(new Date(lastExecutionTime)));
	}
}
