/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.indikator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterObserver;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.StauObjektVerwaltung;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StoerfallIndikator;

/**
 * Verwaltung der auf einer Straße nacheinander folgenden Störfallindikatoren.
 * <br>
 * Die Störfallindikatoren einer Stra&szlig;e k&ouml;nnen theoretisch in einer
 * Richtung und einer Gegenrichtung verbunden werden. In der Praxis ist das in
 * der Regel nicht möglich, da das Netz unvollständig beschrieben ist. Eine
 * Straße wird deshalb in zusammenhängende Ketten von Störfallindikatoren
 * zerlegt, die als Fahrtrichtung betrachtet werden und für jeweils eine Analyse
 * der Staus und die Prognose des Verlaufs derselben erfolgt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class FahrtRichtung implements IFahrtRichtung {

	/** die Liste der Indikatoren, die die Fahrtrichtung bilden. */
	private final ArrayList<ISobIndikator> indikatoren = new ArrayList<>();

	/** die Positionen der Störfallindikatoren wurden bereits bestimmt ? */
	private boolean positioniert;

	/** der Name der Fahrtrichtung. */
	private final String name;

	/** die Straße, auf der die enthaltenen St&ouml;rfallindikatoren liegen. */
	private final SobStrasse strasse;

	/**
	 * der St&ouml;rfallindikator, mit der die Verkettung auf der Stra&szlig;e
	 * begonnen wurde.
	 */
	private final StoerfallIndikator start;

	/** die Staus, die für diese Fahrtrichtung ermittelt wurden. */
	private final StauObjektVerwaltung staus;

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt eine Instanz einer Fahrtrichtung.
	 *
	 * @param strasse
	 *            die Strasse, zu der die Fahrtrichtung gehört
	 * @param name
	 *            der Name
	 * @param start
	 *            der Störfallindikator, mit der die Suche nach verketteten
	 *            Indikatoren begonnen hat
	 */
	public FahrtRichtung(final SobStrasse strasse, final String name, final StoerfallIndikator start) {
		this.strasse = strasse;
		if (name == null) {
			this.name = "namenlos";
		} else {
			this.name = name;
		}
		this.start = start;
		staus = new StauObjektVerwaltung(this);
	}

	@Override
	public final void add(final int index, final ISobIndikator element) {
		positioniert = false;
		indikatoren.add(index, element);
	}

	@Override
	public final boolean add(final ISobIndikator o) {
		positioniert = false;
		return indikatoren.add(o);
	}

	@Override
	public final boolean addAll(final Collection<? extends ISobIndikator> c) {
		positioniert = false;
		return indikatoren.addAll(c);
	}

	@Override
	public final boolean addAll(final int index, final Collection<? extends ISobIndikator> c) {
		positioniert = false;
		return indikatoren.addAll(index, c);
	}

	@Override
	public final void aktualisiereStaus(final long zeitpunkt) {
		staus.verzoegerteTrennung(zeitpunkt);
		staus.prognoseVerlaengerungBerechnen();
		for (final ISobIndikator ind : indikatoren) {
			if (ind.isGeaendert()) {
				switch (ind.getSituation()) {
				case STAU:
					staus.neuerStauBereich(ind);
					break;
				case UNDEFINIERT:
				case VERKEHR:
					staus.neuerStaufreierBereich(ind, zeitpunkt);
					break;
				default:
					break;
				}
				ind.setGeaendert(false);
			}
		}
		staus.zusammenfassen();
		staus.prognoseVerlaengerungPruefen();
	}

	/**
	 * ermittelt die Anzahl der innerhalb der Fahrtrichtung verwendeten
	 * Indikatoren.
	 *
	 * @return die Anzahl
	 */
	public int anzahlIndikatoren() {
		return indikatoren.size();
	}

	@Override
	public final void berechnePositionen() {
		if (!positioniert) {
			positioniert = true;
			double position = 0;
			ISobIndikator segment = null;
			for (final ISobIndikator indikator : indikatoren) {
				if (indikator instanceof SobAeusseresStrassenSegment) {
					// if (segment != null) {
					// position += segment.getLaenge();
					// }
					segment = indikator;
					indikator.setPosition(position);
				} else if (indikator instanceof SobInneresStrassenSegment) {
					// if (segment != null) {
					// position += segment.getLaenge();
					// }
					segment = indikator;
					indikator.setPosition(position);
				} else if (indikator instanceof SobStrassenTeilSegment) {
					indikator.setPosition(position);
					position += indikator.getLaenge();
				} else if (indikator instanceof SobStrassenKnoten) {
					indikator.setPosition(position);
				} else if (indikator instanceof SobMessQuerschnitt) {
					if (segment != null) {
						indikator.setPosition((segment.getPosition() + segment.getLaenge())
								- ((SobMessQuerschnitt) indikator).getIndikator().getStrassenSegmentOffset());
					} else {
						indikator.setPosition(0);
					}
				}
			}
		}
	}

	@Override
	public final void clear() {
		positioniert = false;
		indikatoren.clear();
	}

	/**
	 * ermittelt, ob der übergebene Indikator in der Fahrtrichtung enthalten
	 * ist.
	 *
	 * @param indikator
	 *            der Indikator
	 * @return <code>true</code>, wenn er Bestandteil der Fahrtrichtung ist
	 */
	public boolean contains(final ISobIndikator indikator) {
		return indikatoren.contains(indikator);
	}

	/**
	 * liefert den Offset des {@link ISobIndikator}, der dem übergebenen
	 * Störfallindikator entspricht.
	 *
	 * @param indikator
	 *            der Indikator, dessen Index gesucht wird
	 * @return der Index
	 */
	public int findOffset(final StoerfallIndikator indikator) {
		int result = 0;

		for (final ISobIndikator sobIndikator : indikatoren) {
			if (sobIndikator.getIndikator().equals(indikator)) {
				break;
			}
			result++;
		}

		return result;
	}

	@Override
	public final List<ISobIndikator> getElemente() {
		return indikatoren;
	}

	@Override
	public final String getFullName() {
		return getName() + " (" + getStart().getName() + ")";
	}

	/**
	 * ermittelt den Indikator an der übergebenen Position der Liste der
	 * Indikatoren der Fahrtrichtung.
	 *
	 * @param index
	 *            die Position
	 * @return den Indikator
	 */
	public ISobIndikator getIndikator(final int index) {
		return indikatoren.get(index);
	}

	@Override
	public final double getMaxPrognoseLaenge(final double stauBeginn, final double ende) {
		double result = 0;
		int kreuze = 0;
		int dreiecke = 0;
		int as = 0;

		for (final ISobIndikator ind : indikatoren) {
			final double position = ind.getPosition();
			if (position > ende) {
				break;
			}
			if (position >= stauBeginn) {
				if (ind instanceof SobStrassenKnoten) {
					switch (((SobStrassenKnoten) ind).getIndikator().getKnotenTyp()) {
					case AUTOBAHNANSCHLUSS:
						as++;
						break;
					case AUTOBAHNDREIECK:
						dreiecke++;
						break;
					case AUTOBAHNKREUZ:
						kreuze++;
						break;
					case AUTOBAHNENDE:
					case SONSTIG:
					default:
						break;
					}
				}
			}
		}

		final ISobParameterObserver param = strasse.getParameter();
		if (param != null) {
			result = param.getPrognoseA3() - ((double) param.getPrognoseA1() * (kreuze + dreiecke))
					- ((double) param.getPrognoseA2() * as);
		}
		return Math.max(0, result);
	}

	@Override
	public final String getName() {
		return name;
	}

	@Override
	public final StoerfallIndikator getStart() {
		return start;
	}

	@Override
	public final StauObjektVerwaltung getStaus() {
		return staus;
	}

	@Override
	public final ISobStrasse getStrasse() {
		return strasse;
	}

	/**
	 * ermittelt, den Offset des übergebenen Parameters innerhalb der Liste der
	 * Indikatoren der Fahrtrichtung.
	 *
	 * @param indikator
	 *            der Indikator
	 * @return den Index
	 */
	public int indexOf(final ISobIndikator indikator) {
		return indikatoren.indexOf(indikator);
	}

	@Override
	public final ISobIndikator remove(final int index) {
		positioniert = false;
		return indikatoren.remove(index);
	}

	@Override
	public final boolean remove(final Object o) {
		positioniert = false;
		return indikatoren.remove(o);
	}

	@Override
	public final ISobIndikator set(final int index, final ISobIndikator element) {
		positioniert = false;
		return indikatoren.set(index, element);
	}
}
