/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.stau;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.anastau.SvaMessage;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.IFahrtRichtung;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.ISobIndikator;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobMessQuerschnitt;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStoerfallSituation;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenKnoten;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenSegment;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenTeilSegment;
import de.bsvrz.ibv.anastau.progstau.SvpPrognoseDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.DaVKonstanten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenKnoten;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;

/**
 * Die Repräsentation eines Staus innerhalb der Stauverlaufsanalyse.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class Stau implements IStau, ClientSenderInterface {

	/**
	 * Klasse zur Sortierung von Indikatoren innerhalb einer Fahrtrichtung nach
	 * ihrer Position.
	 */
	public static class IndikatorSorter implements Comparator<ISobIndikator> {

		@Override
		public int compare(final ISobIndikator ind0, final ISobIndikator ind1) {
			int result = 0;
			if (ind0.getPosition() < ind1.getPosition()) {
				result = -1;
			} else if (ind0.getPosition() > ind1.getPosition()) {
				result = 1;
			}
			return result;
		}

	}

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * globaler Zähler zum Bilden einer eindeutigen PID für Stauobjekte.
	 */
	private static long pidCounter = System.currentTimeMillis();

	/**
	 * das Systemobjekt, mit dem der Stau innerhalb des Datenverteilers
	 * repräsentiert wird.
	 */
	private SystemObject obj;

	/**
	 * der Startzeitpunkt des Staus.
	 */
	private final long startZeitpunkt;

	/**
	 * die Fahrtrichtung f&uuml;r die der Stau ermittelt wurde.
	 */
	private final IFahrtRichtung richtung;

	/**
	 * die Indikatoren, die den Stau "verursachen".
	 */
	private final Set<ISobIndikator> indikatoren = new HashSet<>();

	/**
	 * die Liste der Indikatoren, die eine Aufspaltung des Stauobjekts
	 * verursacht, wenn die parametrierte Verzögerungszeit überschritten wurde,
	 * inklusive des Zeitpunkts der Statusänderung des Indikators.
	 */
	private final Map<ISobIndikator, Long> trenner = new HashMap<>();

	/**
	 * die Liste der letzten Prognosedaten für den Stau.
	 */
	private SvpPrognoseDaten prognose;

	/**
	 * Indikator, ob eine Senke zum Empfang der Staueigenschaften aktiv ist. Die
	 * Eigenschaften eines Staus werden als Parameter gespeichert, deshalb ist
	 * die Senke in der Regel die Parametrierung des Datenverteilersystems.
	 */
	private boolean senkeAktiv;

	/** Mutex für das Flag {@link #senkeAktiv}. */
	private final Object senkeAktivMutex = new Object();

	/**
	 * der Datensatz mit den Eigenschaften des Staus.
	 */
	private ResultData eigenschaftenDaten;

	/**
	 * die Datenbeschreibung für den Versand der Eigenschaften eines Staus.
	 */
	private DataDescription eigenschaftenDesc;

	/** der Wert der letzten prognostizierten Stauverlaengerung. */
	private double prognoseVerlaengerung;

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt ein neues Stauobjekt für die angegebene
	 * Fahrtrichtung.
	 *
	 * @param richtung
	 *            die Fahrtrichtung
	 */
	public Stau(final IFahrtRichtung richtung) {
		this.richtung = richtung;
		startZeitpunkt = System.currentTimeMillis();
	}

	/**
	 * fügt dem Stauobjekt einen Indikator hinzu.<br>
	 * Der Indikator muss sich im Status <code>STAU</code> befinden.
	 *
	 * @param indikator
	 *            der hinzufügende Indikator
	 */
	public final void addIndikator(final ISobIndikator indikator) {
		if (!indikator.getSituation().equals(SobStoerfallSituation.STAU)) {
			LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(indikator.getIndikator().getSystemObject()),
					SvaMessage.WarningIndikatorOhneStau, this, indikator);
		} else {
			indikatoren.add(indikator);
			trenner.remove(indikator);
		}
	}

	/**
	 * fügt dem Stauobjekt eine Liste von Indikatoren hinzu.<br>
	 * Der Indikatoren müssen sich im Status <code>STAU</code> befinden.
	 *
	 * @param indikatorliste
	 *            die Liste der hinzuzufügenden Indikatoren.
	 */
	public final void addIndikatoren(final Collection<ISobIndikator> indikatorliste) {
		for (final ISobIndikator indikator : indikatorliste) {
			indikatoren.add(indikator);
		}
	}

	@Override
	public void aktualisierePrognoseVerlaengerung() {

		if ((prognose != null) && (prognose.size() > 1)) {
			prognoseVerlaengerung = prognose.get(1).getLaenge() - getLaenge();
		} else {
			prognoseVerlaengerung = 0;
		}

		final double indikatorLaenge = getLaenge();
		final double start = getStrassenStart();
		final double strassenEnde = getStrassenEnde();
		final List<ISobIndikator> indList = new ArrayList<>(getIndikatoren());
		ISobIndikator letzterIndikator = indList.get(indList.size() - 1);
		letzterIndikator = letzterIndikator.getNext();
		while (letzterIndikator != null) {
			if (letzterIndikator.getStauBeginn() > strassenEnde) {
				letzterIndikator = null;
				break;
			}
			if (letzterIndikator.isFreierVerkehr()) {
				break;
			}
			letzterIndikator = letzterIndikator.getNext();
		}

		if (letzterIndikator != null) {
			prognoseVerlaengerung = letzterIndikator.getStauBeginn() - start - indikatorLaenge;
		}
	}

	@Override
	public final long anzahlFahrStreifen(final long offset) {
		long result = 2;
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind.getIndikator() instanceof StrassenTeilSegment) {
				if ((ind.getPosition() <= offset) && ((ind.getPosition() + ind.getLaenge()) >= offset)) {
					result = Math.max(2, ((StrassenTeilSegment) ind.getIndikator()).getAnzahlFahrstreifen());
					break;
				}
			}
		}

		return result;
	}

	/**
	 * Die Funktion verarbeitet den Status der Senke für die Eigenschaften des
	 * Stauobjekts. Der Datensatz mit den Eigenschaften wird bei Bedarf
	 * versendet.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription, byte)
	 */
	@Override
	public final void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		if (dataDescription.getAttributeGroup().equals(eigenschaftenDesc.getAttributeGroup())) {
			if (state == ClientSenderInterface.START_SENDING) {
				synchronized (senkeAktivMutex) {
					senkeAktiv = true;
					if (eigenschaftenDaten != null) {
						try {
							StauverlaufsAnalyse.getDav().sendData(eigenschaftenDaten);
						} catch (final DataNotSubscribedException e) {
							// Sollte nicht auftreten, da erst nach Bestätigung
							// der Anmeldung gesendet wird.
							LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(object),
									SvaMessage.ErrorSendeStauEigenschaften, e.getMessage());
						} catch (final SendSubscriptionNotConfirmed e) {
							// Sollte nicht auftreten, da erst nach Bestätigung
							// der Anmeldung gesendet wird.
							LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(object),
									SvaMessage.ErrorSendeStauEigenschaften, e.getMessage());
						}
					}
				}

			} else {
				senkeAktiv = false;
			}
		}
	}

	/**
	 * entfernt den Indikator aus dem Stauobjekt.
	 *
	 * @param indikator
	 *            der zu entfernende Indikator
	 */
	public void entferneIndikator(final ISobIndikator indikator) {
		indikatoren.remove(indikator);
		trenner.remove(indikator);
	}

	/**
	 * Die Funktion entfernt einen Stau vom betroffenen Verkehsmodellnetz.
	 *
	 * @see de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau#entfernen()
	 */
	@Override
	public final void entfernen() {
		if (obj != null) {

			final ClientDavInterface dav = StauverlaufsAnalyse.getDav();
			dav.unsubscribeSender(this, obj, eigenschaftenDesc);
			dav.unsubscribeSender(this, obj, SvpPrognoseDaten.getDataDesc());

			try {
				final VerkehrModellNetz netz = richtung.getStrasse().getNetz();
				if (netz != null) {
					netz.stauEntfernen(obj);
				}
				obj.invalidate();
			} catch (final ConfigurationChangeException e) {
				LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(obj), SvaMessage.ErrorStauObjektLoeschen,
						e.getMessage());
			}
		}
	}

	/**
	 * entfernt einen als Stautrenner vorgemerktemn Indikator von dem
	 * Stauobjekt.
	 *
	 * @param indikator
	 *            der Indikator
	 */
	public void entferneTrenner(final ISobIndikator indikator) {
		trenner.remove(indikator);
	}

	@Override
	public Collection<SobStrassenKnoten> ermittleAsAmAnfang(final MessQuerschnittAllgemein mq) {
		final Collection<SobStrassenKnoten> result = new ArrayList<>();
		final double stauAnfang = getStrassenStart();
		boolean mqGefunden = false;

		if (mq != null) {
			for (final ISobIndikator indikator : richtung.getElemente()) {
				if (indikator.getPosition() > stauAnfang) {
					break;
				} else if (indikator instanceof SobMessQuerschnitt) {
					if (mq.equals(indikator.getIndikator())) {
						mqGefunden = true;
					}
				} else if (indikator instanceof SobStrassenKnoten) {
					if (mqGefunden) {
						result.add((SobStrassenKnoten) indikator);
					}
				}
			}
		}

		return result;
	}

	@Override
	public Collection<SobStrassenKnoten> ermittleAsAmEnde(final MessQuerschnittAllgemein mq) {

		final Collection<SobStrassenKnoten> result = new ArrayList<>();
		final double stauEnde = getStrassenEnde();

		if (mq != null) {
			for (final ISobIndikator indikator : richtung.getElemente()) {
				if (indikator.getPosition() > stauEnde) {
					if (indikator instanceof SobMessQuerschnitt) {
						if (mq.equals(indikator.getIndikator())) {
							break;
						}
					}
					if (indikator instanceof SobStrassenKnoten) {
						result.add((SobStrassenKnoten) indikator);
					}
				}
			}
		}

		return result;
	}

	/**
	 * ermitteln des Offsets des Stauendes relativ zum Ende des letzten
	 * Straßensegments, das am Stau beteiligt ist.
	 *
	 * @param segmente
	 *            die sortierte Liste der am Stau beteiligten Straßensegmente.
	 * @return den Offset
	 */
	private int getAnfangsOffset(final List<StrassenSegment> segmente) {
		int result = 0;

		if (segmente.size() > 0) {
			final StrassenSegment segment = segmente.get(0);
			for (final ISobIndikator ind : richtung.getElemente()) {
				if (ind.getIndikator().equals(segment)) {
					final double ende = getStrassenEnde();
					final double laenge = ind.getLaenge();
					final double position = ind.getPosition();
					result = Math.max(0, (int) (laenge - (ende - position)));
					break;
				}
			}
		}

		return result;
	}

	/**
	 * ermitteln des Offsets des Staubeginns relativ zum ersten Straßensegment,
	 * das am Stau beteiligt ist.
	 *
	 * @param segmente
	 *            die sortierte Liste der am Stau beteiligten Straßensegmente.
	 * @return den Offset
	 */
	private int getEndeOffset(final List<StrassenSegment> segmente) {
		int result = 0;

		if (segmente.size() > 0) {
			final StrassenSegment segment = segmente.get(segmente.size() - 1);
			for (final ISobIndikator ind : richtung.getElemente()) {
				if (ind.getIndikator().equals(segment)) {
					final double anfang = getStrassenStart();
					final double position = ind.getPosition();
					final double laenge = ind.getLaenge();
					result = Math.max(0, (int) ((position + laenge) - anfang));

					break;
				}
			}
		}

		return result;
	}

	/**
	 * liefert die Liste der am Stau beteiligten Indikatoren.<br>
	 * Die Liste enthält alle Indikatoren, deren Status auf <code>STAU</code>
	 * gesetzt ist.
	 *
	 * @return die Liste
	 */
	public final List<ISobIndikator> getIndikatoren() {
		final List<ISobIndikator> result = new ArrayList<>(indikatoren);
		Collections.sort(result, new IndikatorSorter());
		return result;
	}

	@Override
	public Collection<SobStrassenKnoten> getKnotenImStau() {
		final Collection<SobStrassenKnoten> result = new ArrayList<>();
		final double stauBeginn = getStrassenStart();
		final double stauEnde = getStrassenEnde();

		for (final ISobIndikator indikator : richtung.getElemente()) {
			if (indikator.getPosition() > stauEnde) {
				break;
			}
			if (indikator.getPosition() > stauBeginn) {
				if (indikator instanceof SobStrassenKnoten) {
					result.add((SobStrassenKnoten) indikator);
				}
			}
		}

		return result;
	}

	@Override
	public final long getLaenge() {
		double result = 0;
		ISobIndikator anfang = null;
		ISobIndikator ende = null;
		for (final ISobIndikator indikator : indikatoren) {
			if ((anfang == null) || (ende == null)) {
				anfang = indikator;
				ende = indikator;
			} else {
				if (indikator.getPosition() < anfang.getPosition()) {
					anfang = indikator;
				}
				if ((indikator.getPosition() + indikator.getLaenge()) > (ende.getPosition() + ende.getLaenge())) {
					ende = indikator;
				}
			}
		}

		if ((anfang != null) && (ende != null)) {
			// result = (ende.getPosition() + ende.getLaenge())
			// - anfang.getPosition();
			result = ende.getStauEnde() - anfang.getStauBeginn();
		}

		return (long) result;
	}

	@Override
	public SobStrassenTeilSegment getLetztesStsImStau() {
		final double stauAnfang = getStrassenStart();
		final double stauEnde = getStrassenEnde();
		SobStrassenTeilSegment result = null;

		for (final ISobIndikator indikator : richtung.getElemente()) {
			if (indikator.getStauBeginn() >= stauEnde) {
				break;
			}
			if (indikator instanceof SobStrassenTeilSegment) {
				if (indikator.getStauEnde() > stauAnfang) {
					result = (SobStrassenTeilSegment) indikator;
					break;
				}
			}
		}

		return result;
	}

	@Override
	public final SobMessQuerschnitt getMqAmAnfang() {
		SobMessQuerschnitt result = null;
		final double anfang = getStrassenStart();
		final double ende = getStrassenEnde();
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind.getPosition() >= anfang) {
				if (ind instanceof SobMessQuerschnitt) {
					result = (SobMessQuerschnitt) ind;
					break;
				}
			}
			if (ind.getPosition() > ende) {
				break;
			}
		}
		return result;
	}

	@Override
	public final SobMessQuerschnitt getMqAmEnde() {
		SobMessQuerschnitt result = null;
		final double anfang = getStrassenStart();
		final double ende = getStrassenEnde();
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind.getPosition() >= anfang) {
				if (ind instanceof SobMessQuerschnitt) {
					result = (SobMessQuerschnitt) ind;
				}
			}
			if (ind.getPosition() > ende) {
				break;
			}
		}
		return result;
	}

	@Override
	public final SobMessQuerschnitt getMqDanach() {
		SobMessQuerschnitt result = null;
		final double anfang = getStrassenStart();
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind.getPosition() > anfang) {
				break;
			}
			if (ind instanceof SobMessQuerschnitt) {
				result = (SobMessQuerschnitt) ind;
			} else if (ind.getIndikator() instanceof StrassenKnoten) {
				switch (((StrassenKnoten) ind.getIndikator()).getKnotenTyp()) {
				case AUTOBAHNDREIECK:
				case AUTOBAHNKREUZ:
				case AUTOBAHNENDE:
					result = null;
					break;
				case AUTOBAHNANSCHLUSS:
				case SONSTIG:
				default:
					break;
				}
			}
		}
		return result;
	}

	@Override
	public final SobMessQuerschnitt getMqDavor() {
		SobMessQuerschnitt result = null;
		final double ende = getStrassenEnde();
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind.getPosition() > ende) {
				if (ind instanceof SobMessQuerschnitt) {
					result = (SobMessQuerschnitt) ind;
					break;
				} else if (ind.getIndikator() instanceof StrassenKnoten) {
					boolean stop = false;
					switch (((StrassenKnoten) ind.getIndikator()).getKnotenTyp()) {
					case AUTOBAHNDREIECK:
					case AUTOBAHNKREUZ:
					case AUTOBAHNENDE:
						stop = true;
						break;
					case AUTOBAHNANSCHLUSS:
					case SONSTIG:
					default:
						break;
					}
					if (stop) {
						break;
					}
				}
			}
		}
		return result;
	}

	/**
	 * liefert die nächste an einen Stau zu vergebende PID. Die Pid lautet
	 * "sva.stau.&lt;xxx&gt;", wobei "xxx" eine laufende Nummer ist, die beim
	 * Start der Applikation mit dem aktuellen Zeitstempel in Millisekunden
	 * initialisiert wird.
	 *
	 * @return die PID
	 */
	private String getNextPid() {
		Stau.pidCounter++;
		return "sva.stau." + Stau.pidCounter;
	}

	@Override
	public final long getOffset() {
		return (long) getStrassenStart();
	}

	@Override
	public SvpPrognoseDaten getPrognose() {
		return prognose;
	}

	/**
	 * erzeugt einen Datensatz zum versenden der aktuellen Prognosedaten des
	 * Staus.
	 *
	 * @return den Datensatz
	 */
	private ResultData getPrognoseDatenSatz() {

		if (obj == null) {
			throw new IllegalStateException("Stauprognose kann nicht ohne Systemobjekt gespeichert werden.");
		}

		final Data data = prognose.getDatenSatz();
		return new ResultData(obj, SvpPrognoseDaten.getDataDesc(), prognose.getPrognoseZeitPunkt(), data);
	}

	/**
	 * liefert die Länge, um die ein Stauobjekt durch die aktuelle Prognose
	 * verlängert wird.
	 *
	 * @return die Länge
	 */
	double getPrognoseVerlaengerung() {
		return prognoseVerlaengerung;
	}

	@Override
	public final long getStartZeit() {
		return startZeitpunkt;
	}

	/**
	 * ermittelt die Länge der Lücke, die ein ungestauter Indikator innerhalb
	 * eines Stauobjekts öffnen würde.
	 *
	 * @param ind
	 *            der Indikator.
	 * @return den Wert
	 */
	public double getStauLuecke(final ISobIndikator ind) {
		boolean gefunden = false;
		ISobIndikator davor = null;
		ISobIndikator danach = null;

		for (final ISobIndikator indikator : getIndikatoren()) {
			if (indikator.equals(ind)) {
				gefunden = true;
			} else {
				if (gefunden) {
					danach = indikator;
					break;
				}
				davor = indikator;
			}
		}

		double result = 0;
		if (gefunden) {
			result = Double.MAX_VALUE;
			if ((davor != null) && (danach != null)) {
				result = danach.getStauBeginn() - ind.getStauBeginn();
			}
		}

		return result;
	}

	@Override
	public final double getStrassenEnde() {

		final double start = getStrassenStart();
		final double indikatorLaenge = getLaenge();
		final double result = Math.max(0, start + indikatorLaenge + prognoseVerlaengerung);
		return Math.max(start, result);
	}

	@Override
	public final List<StrassenSegment> getStrassenSegmente() {
		final List<StrassenSegment> result = new ArrayList<>();
		final double start = getStrassenStart();
		final double ende = getStrassenEnde();
		for (final ISobIndikator ind : richtung.getElemente()) {
			if (ind instanceof SobStrassenSegment) {
				final double segmentStart = ind.getPosition();
				final double segmentEnde = ind.getPosition() + ind.getLaenge();
				if ((segmentStart >= start) && (segmentStart <= ende)) {
					result.add(0, (StrassenSegment) ind.getIndikator());
				} else if ((segmentEnde >= start) && (segmentEnde <= ende)) {
					result.add(0, (StrassenSegment) ind.getIndikator());
				} else if ((segmentStart < start) && (segmentEnde > ende)) {
					result.add(0, (StrassenSegment) ind.getIndikator());
				}
			}
		}

		return result;
	}

	// /**
	// * fügt einen Störfallindikator in die Liste der Indikatoren ein, die nach
	// * einer parametrierbaren Zeit zur Trennung eines Staus führen können.<br>
	// * Der Indikator wird nur eingetragen, wenn sein Zustand auf Grund der
	// Lage
	// * inerhalb des Staus nicht sofort zur Teilung desselben führt. Als
	// Ergebnis
	// * gibt die Funktion die Liste der neuen Staus zurück, die entweder aus
	// dem
	// * Stau selbst besteht, wenn die Trennung noch nicht vorgenommen wurde
	// oder
	// * aus der Liste der durch die Trennung neu entstandenen Stauobjekte. Wenn
	// * der Indikator der letzte war, der den Stau beschrieben hat, wird eine
	// * leere Liste geliefert.
	// *
	// * @param ind
	// * der Indikator
	// * @return die Liste der resultierenden Staus
	// */
	// public final Collection<Stau> setTrenner(final ISobIndikator ind) {
	// final Collection<Stau> result = new ArrayList<Stau>();
	// final double luecke = getStauLuecke(ind);
	// indikatoren.remove(ind);
	//
	// if (indikatoren.size() > 0) {
	// if (luecke < richtung.getStrasse().getParameter()
	// .getTeilungLaenge()) {
	// if (trenner.containsKey(ind)) {
	// trenner.remove(ind);
	// } else {
	// trenner.put(ind, System.currentTimeMillis());
	// }
	// result.add(this);
	// } else {
	// Collection<ISobIndikator> indList = new ArrayList<ISobIndikator>(
	// indikatoren);
	// indList.remove(ind);
	// indikatoren.clear();
	// for (ISobIndikator indikator : richtung.getElemente()) {
	// if (indikator.equals(ind)) {
	// break;
	// }
	// if (indList.contains(indikator)) {
	// indikatoren.add(indikator);
	// indList.remove(indikator);
	// }
	// }
	// if (indikatoren.size() > 0) {
	// result.add(this);
	// }
	// if (indList.size() > 0) {
	// Stau neuerStau = new Stau(richtung);
	// neuerStau.indikatoren.addAll(indList);
	// result.add(neuerStau);
	// }
	// }
	// }
	//
	// return result;
	// }

	@Override
	public final double getStrassenStart() {
		double result = Double.MAX_VALUE;
		for (final ISobIndikator ind : indikatoren) {
			result = Math.min(result, ind.getStauBeginn());
		}

		return result;
	}

	@Override
	public SobStrassenTeilSegment getStsVorDemStau() {
		final double stauAnfang = getStrassenStart();
		SobStrassenTeilSegment result = null;

		for (final ISobIndikator indikator : richtung.getElemente()) {
			if (indikator instanceof SobStrassenTeilSegment) {
				if (indikator.getStauEnde() <= stauAnfang) {
					result = (SobStrassenTeilSegment) indikator;
				} else {
					break;
				}
			}
		}

		return result;
	}

	/**
	 * liefert die Indikatoren, die potentiell vorgemerkt sind, um ein
	 * Stauobjekt in kleiner Teilobjekte zu teilen.
	 *
	 * @return die Menge der Indikatoren
	 */
	public Map<ISobIndikator, Long> getTrenner() {
		return new HashMap<>(trenner);
	}

	/**
	 * Die Sendesteuerung wird für die Versendung der Eigenschaften des Staus
	 * (als Parameter) unterstützt.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject,
	 *      de.bsvrz.dav.daf.main.DataDescription)
	 */
	@Override
	public final boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		boolean result = false;
		if (dataDescription.getAttributeGroup().equals(eigenschaftenDesc.getAttributeGroup())) {
			result = true;
		}
		return result;
	}

	@Override
	public void publizierePrognoseDaten() {
		if (obj != null) {
			if (prognose != null) {
				try {
					final ResultData daten = getPrognoseDatenSatz();
					if ((daten.getData() != null) && daten.getData().isDefined()) {
						StauverlaufsAnalyse.getDav().sendData(daten);
					} else {
						Debug.getLogger().warning(
								"Der unvollständige Prognosedatensatz konnte nicht versendet werden: " + daten);
					}
				} catch (final DataNotSubscribedException e) {
					throw new IllegalStateException("Prognosedaten lassen sich nicht versenden", e);
				} catch (final SendSubscriptionNotConfirmed e) {
					throw new IllegalStateException("Prognosedaten lassen sich nicht versenden", e);
				}
			}
		}
	}

	@Override
	public final void setPrognose(final SvpPrognoseDaten prognose) {
		this.prognose = prognose;

	}

	/**
	 * setzt die aktuell gültige Länge, um die ein Stauobjekt durch die Prognose
	 * verlängert wird.
	 *
	 * @param laenge
	 *            der Wert für die Verlängerung
	 */
	void setPrognoseVerlaengerung(final double laenge) {
		this.prognoseVerlaengerung = laenge;
	}

	/**
	 * registriert einen Störfallindikator, der zu einer potentiellen eventuell
	 * verzögerten Trennung des Stauobjekt führen kann.
	 *
	 * @param ind
	 *            der Indikator
	 * @param zeitpunkt
	 *            der Zeitpunkt, zu dem der Indikator eingetragen wird
	 */
	public void setTrenner(final ISobIndikator ind, final long zeitpunkt) {
		trenner.put(ind, zeitpunkt);
	}

	@Override
	public final void sichern(final long zeitpunkt) {

		Stau.LOGGER.finer("Sichere Stau auf: " + richtung.getStrasse().getStrasse().getName() + " von "
				+ getStrassenStart() + " m bis " + getStrassenEnde() + " m");

		final ClientDavInterface dav = StauverlaufsAnalyse.getDav();
		final DataModel model = dav.getDataModel();
		eigenschaftenDesc = new DataDescription(model.getAttributeGroup(StauDefs.PID_ATG_SITUATIONS_EIGENSCHAFTEN),
				model.getAspect(DaVKonstanten.ASP_PARAMETER_VORGABE));

		if (obj == null) {
			Stau.LOGGER.finest("Neues Stauobjekt anlegen");
			try {
				final VerkehrModellNetz netz = richtung.getStrasse().getNetz();

				obj = DynamischeObjekte.getInstanz(StauverlaufsAnalyse.getDav())
						.erzeugeObjekt((DynamicObjectType) model.getType(StauDefs.PID_TYP_STAU), null, getNextPid());

				dav.subscribeSender(this, obj, eigenschaftenDesc, SenderRole.sender());

				dav.subscribeSender(this, obj, SvpPrognoseDaten.getDataDesc(), SenderRole.source());
				if (prognose != null) {
					final ResultData daten = getPrognoseDatenSatz();
					if ((daten.getData() != null) && daten.getData().isDefined()) {
						StauverlaufsAnalyse.getDav().sendData(daten);
					} else {
						Debug.getLogger().warning(
								"Der unvollständige Prognosedatensatz konnte nicht versendet werden: " + daten);
					}
					dav.sendData(daten);
				}
				Stau.LOGGER.finest("Neues Stauobjekt: " + obj.getNameOrPidOrId());
				if (netz != null) {
					netz.stauHinzufuegen(obj);
				}
			} catch (final OneSubscriptionPerSendData e) {
				LogTools.log(Stau.LOGGER, SvaMessage.ErrorStauObjektAnlegen, new BetriebsmeldungDaten(obj),
						e.getMessage());
				throw new RuntimeException(
						MessageFormat.format(SvaMessage.ErrorStauObjektAnlegen.name(), e.getMessage()));
			} catch (final DataNotSubscribedException e) {
				LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(obj), SvaMessage.ErrorStauObjektAnlegen,
						e.getMessage());
				throw new RuntimeException(
						MessageFormat.format(SvaMessage.ErrorStauObjektAnlegen.name(), e.getMessage()));
			} catch (final SendSubscriptionNotConfirmed e) {
				LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(obj), SvaMessage.ErrorStauObjektAnlegen,
						e.getMessage());
				throw new RuntimeException(
						MessageFormat.format(SvaMessage.ErrorStauObjektAnlegen.name(), e.getMessage()));
			} catch (final DynObjektException e) {
				LogTools.log(Stau.LOGGER, new BetriebsmeldungDaten(obj), SvaMessage.ErrorStauObjektAnlegen,
						e.getMessage());
				throw new RuntimeException(
						MessageFormat.format(SvaMessage.ErrorStauObjektAnlegen.name(), e.getMessage()));
			}
		}

		if (obj == null) {
			throw new IllegalStateException(
					"Für das zu sicherende Stauobjekt besteht kein Systemobjekt in der Datenverteilerkonfiguration");
		}

		final Data data = StauverlaufsAnalyse.getDav().createData(eigenschaftenDesc.getAttributeGroup());
		data.getTimeValue(StauDefs.ATT_EIGENSCHAFTEN_START_ZEIT).setMillis(startZeitpunkt);
		data.getTimeValue(StauDefs.ATT_EIGENSCHAFTEN_DAUER).setMillis(0);

		final List<StrassenSegment> segmente = getStrassenSegmente();
		final Data.Array array = data.getArray(StauDefs.ATT_EIGENSCHAFTEN_STRASSEN_SEGMENT);
		array.setLength(segmente.size());
		for (int idx = 0; idx < segmente.size(); idx++) {
			array.getReferenceValue(idx).setSystemObject(segmente.get(idx).getSystemObject());
		}

		data.getScaledValue(StauDefs.ATT_EIGENSCHAFTEN_START_OFFSET).set(getAnfangsOffset(segmente));
		data.getScaledValue(StauDefs.ATT_EIGENSCHAFTEN_END_OFFSET).set(getEndeOffset(segmente));
		data.getScaledValue(StauDefs.ATT_EIGENSCHAFTEN_LAENGE).set(getLaenge());

		synchronized (senkeAktivMutex) {
			eigenschaftenDaten = new ResultData(obj, eigenschaftenDesc, zeitpunkt, data);
			if (senkeAktiv) {

				try {
					StauverlaufsAnalyse.getDav().sendData(eigenschaftenDaten);
				} catch (final DataNotSubscribedException e) {
					throw new IllegalStateException("Staueigenschaften konnten nicht versendet werden", e);
				} catch (final SendSubscriptionNotConfirmed e) {
					throw new IllegalStateException("Staueigenschaften konnten nicht versendet werden", e);
				}
			}
		}
	}

	@Override
	public String toString() {
		String result = null;
		if (obj == null) {
			if (richtung != null) {
				result = "ungesicherter Stau auf: " + richtung.getStrasse().getStrasse().getName() + " von "
						+ getStrassenStart() + " m bis " + getStrassenEnde() + " m";
			} else {
				result = "nicht näher definierter ungesicherter Stau";
			}
		} else {
			result = obj.toString();
		}
		return result;
	}
}
